TOPIC_LIST_COLLECTIONS: bytes = b"lc"
TOPIC_LIST_COLLECTION_STREAMS: bytes = b"lcs"
TOPIC_REPLAY_COLLECTION_STREAM: bytes = b"rcs"
TOPIC_PUBLISH_COLLECTION_STREAM: bytes = b"pcs"
TOPIC_LIST_LIVE_STREAMS: bytes = b"lls"
TOPIC_READ_LIVE_STREAM: bytes = b"rls"
TOPIC_STOP_TASK: bytes = b"st"
